@@ -7,7 +7,7 @@
 # Copyright (c) 1992-1994 The Regents of the University of California.
 # Copyright (c) 1994 Sun Microsystems, Inc.
 # perl/Tk version:
-# Copyright (c) 1995-2000 Nick Ing-Simmons
+# Copyright (c) 1995-2004 Nick Ing-Simmons
 # Copyright (c) 1999 Greg London
 #
 # See the file "license.terms" for information on usage and redistribution
@@ -20,7 +20,8 @@ use strict;
 use Text::Tabs;
 
 use vars qw($VERSION);
-$VERSION = '4.013'; # $Id: //depot/Tkutf8/Text/Text.pm#13 $
+#$VERSION = sprintf '4.%03d', q$Revision: #24 $ =~ /\D(\d+)\s*$/;
+$VERSION = '4.025';
 
 use Tk qw(Ev $XS_VERSION);
 use base  qw(Tk::Clipboard Tk::Widget);
@@ -33,7 +34,7 @@ sub Tk_cmd { \&Tk::text }
 
 sub Tk::Widget::ScrlText { shift->Scrolled('Text' => @_) }
 
-Tk::Methods('bbox','compare','debug','delete','dlineinfo','dump',
+Tk::Methods('bbox','compare','debug','delete','dlineinfo','dump','edit',
             'get','image','index','insert','mark','scan','search',
             'see','tag','window','xview','yview');
 
@@ -45,6 +46,7 @@ use Tk::Submethods ( 'mark'   => [qw(gravity names next previous set unset)],
 		     'image'  => [qw(cget configure create names)],
 		     'xview'  => [qw(moveto scroll)],
 		     'yview'  => [qw(moveto scroll)],
+                     'edit'   => [qw(modified redo reset separator undo)],
 		     );
 
 sub Tag;
@@ -145,6 +147,8 @@ sub bindRdOnly
  $mw->YMouseWheelBind($class);
  $mw->XMouseWheelBind($class);
 
+ $mw->MouseWheelBind($class);
+
  return $class;
 }
 
@@ -218,7 +222,7 @@ sub ClassInit
    $mw->bind($class,'<Control-h>','deleteBefore');
    $mw->bind($class,'<ButtonRelease-2>','ButtonRelease2');
   }
- $Tk::prevPos = undef;
+#JD# $Tk::prevPos = undef;
  return $class;
 }
 
@@ -272,7 +276,12 @@ sub ButtonRelease2
  my $Ev = $w->XEvent;
  if (!$Tk::mouseMoved)
   {
-   Tk::catch { $w->insert($Ev->xy,$w->SelectionGet) }
+   Tk::catch
+    {
+     $w->mark('set','insert',$Ev->xy);
+     $w->insert($Ev->xy,$w->SelectionGet);
+     $w->focus if ($w->cget('-state') eq "normal");
+    }
   }
 }
 
@@ -482,7 +491,7 @@ sub SetCursor
  $pos = 'end - 1 chars' if $w->compare($pos,'==','end');
  $w->markSet('insert',$pos);
  $w->unselectAll;
- $w->see('insert')
+ $w->see('insert');
 }
 # KeySelect
 # This procedure is invoked when stroking out selections using the
@@ -631,7 +640,7 @@ sub ToggleInsertMode
 sub InsertKeypress
 {
  my ($w,$char)=@_;
- return unless length $char;
+ return unless length($char);
  if ($w->OverstrikeMode)
   {
    my $current=$w->get('insert');
@@ -998,7 +1007,7 @@ sub findandreplacepopup
  my ($w,$find_only)=@_;
 
  my $pop = $w->Toplevel;
- $pop->transient($w->MainWindow);
+ $pop->transient($w->toplevel);
  if ($find_only)
   { $pop->title("Find"); }
  else
@@ -1010,31 +1019,31 @@ sub findandreplacepopup
  my $direction = '-forward';
  $frame->Radiobutton(
   -variable => \$direction,
-  -text => '-forward',-value => '-forward' )
+  -text => 'forward',-value => '-forward' )
   ->grid(-row=> 2, -column=>1, -padx=> 20, -sticky => 'nw');
  $frame->Radiobutton(
   -variable => \$direction,
-  -text => '-backward',-value => '-backward' )
+  -text => 'backward',-value => '-backward' )
   ->grid(-row=> 3, -column=>1, -padx=> 20, -sticky => 'nw');
 
  $frame->Label(-text=>"Mode:")
   ->grid(-row=> 1, -column=>2, -padx=> 20, -sticky => 'nw');
  my $mode = '-exact';
  $frame->Radiobutton(
-  -variable => \$mode, -text => '-exact',-value => '-exact' )
+  -variable => \$mode, -text => 'exact',-value => '-exact' )
   ->grid(-row=> 2, -column=>2, -padx=> 20, -sticky => 'nw');
  $frame->Radiobutton(
-  -variable => \$mode, -text => '-regexp',-value => '-regexp' )
+  -variable => \$mode, -text => 'regexp',-value => '-regexp' )
   ->grid(-row=> 3, -column=>2, -padx=> 20, -sticky => 'nw');
 
  $frame->Label(-text=>"Case:")
   ->grid(-row=> 1, -column=>3, -padx=> 20, -sticky => 'nw');
  my $case = '-case';
  $frame->Radiobutton(
-  -variable => \$case, -text => '-case',-value => '-case' )
+  -variable => \$case, -text => 'case',-value => '-case' )
   ->grid(-row=> 2, -column=>3, -padx=> 20, -sticky => 'nw');
  $frame->Radiobutton(
-  -variable => \$case, -text => '-nocase',-value => '-nocase' )
+  -variable => \$case, -text => 'nocase',-value => '-nocase' )
   ->grid(-row=> 3, -column=>3, -padx=> 20, -sticky => 'nw');
 
  ######################################################
@@ -1107,6 +1116,9 @@ sub findandreplacepopup
   $find_entry->bind("<Return>" => [$button_find, 'invoke']);
   $find_entry->bind("<Escape>" => [$button_cancel, 'invoke']);
 
+ $find_entry->bind("<Return>" => [$button_find, 'invoke']);
+ $find_entry->bind("<Escape>" => [$button_cancel, 'invoke']);
+
  $pop->resizable('yes','no');
  return $pop;
 }
@@ -1152,63 +1164,101 @@ sub Insert
 }
 
 # UpDownLine --
-# Returns the index of the character one line above or below the
+# Returns the index of the character one *display* line above or below the
 # insertion cursor. There are two tricky things here. First,
 # we want to maintain the original column across repeated operations,
 # even though some lines that will get passed through do not have
 # enough characters to cover the original column. Second, do not
 # try to scroll past the beginning or end of the text.
 #
+# This may have some weirdness associated with a proportional font. Ie.
+# the insertion cursor will zigzag up or down according to the width of
+# the character at destination.
+#
 # Arguments:
 # w - The text window in which the cursor is to move.
 # n - The number of lines to move: -1 for up one line,
 # +1 for down one line.
-sub UpDownLine_old
-{
- my ($w,$n) = @_;
- my $i = $w->index('insert');
- my ($line,$char) = split(/\./,$i);
- if (!defined($Tk::prevPos) || $Tk::prevPos ne $i)
-  {
-   $Tk::char = $char
-  }
- my $new = $w->index($line+$n . '.' . $Tk::char);
- if ($w->compare($new,'==','end') || $w->compare($new,'==','insert linestart'))
-  {
-   $new = $i
-  }
- $Tk::prevPos = $new;
- return $new;
-}
-
 sub UpDownLine
 {
- my ($w,$n) = @_;
- my $i = $w->index('insert');
- my ($line,$char) = split(/\./,$i);
- my $string = $w->get($line.'.0', $i);
+my ($w,$n) = @_;
+$w->see('insert');
+my $i = $w->index('insert');
 
- $string = expand($string);
- $char=length($string);
- $line += $n;
+my ($line,$char) = split(/\./,$i);
 
- $string = $w->get($line.'.0', $line.'.0 lineend');
- $string = expand($string);
- $string = substr($string, 0, $char);
+my $testX; #used to check the "new" position
+my $testY; #used to check the "new" position
 
- $string = unexpand($string);
- $char = length($string);
+(my $bx, my $by, my $bw, my $bh) = $w->bbox($i);
+(my $lx, my $ly, my $lw, my $lh) = $w->dlineinfo($i);
 
- my $new = $w->index($line . '.' . $char);
- if ($w->compare($new,'==','end') || $w->compare($new,'==','insert linestart'))
+if ( ($n == -1) and ($by <= $bh) )
   {
-   $new = $i
+   #On first display line.. so scroll up and recalculate..
+   $w->yview('scroll', -1, 'units');
+   unless (($w->yview)[0]) {
+     #first line of entire text - keep same position.
+     return $i;
+   }
+   ($bx, $by, $bw, $bh) = $w->bbox($i);
+   ($lx, $ly, $lw, $lh) = $w->dlineinfo($i);
+  }
+elsif ( ($n == 1) and
+         ($ly + $lh) > ( $w->height - 2*$w->cget(-bd) - 2*$w->cget(-highlightthickness) ) )
+  {
+   #On last display line.. so scroll down and recalculate..
+   $w->yview('scroll', 1, 'units');
+   ($bx, $by, $bw, $bh) = $w->bbox($i);
+   ($lx, $ly, $lw, $lh) = $w->dlineinfo($i);
   }
- $Tk::prevPos = $new;
- $Tk::char = $char;
- return $new;
-}
 
+# Calculate the vertical position of the next display line
+my $Yoffset = 0;
+$Yoffset = $by - $ly + 1 if ($n== -1);
+$Yoffset = $ly + $lh + 1 - $by if ($n == 1);
+$Yoffset*=$n;
+$testY = $by + $Yoffset;
+
+# Save the original 'x' position of the insert cursor if:
+# 1. This is the first time through -- or --
+# 2. The insert cursor position has changed from the previous
+#    time the up or down key was pressed -- or --
+# 3. The cursor has reached the beginning or end of the widget.
+
+if (not defined $w->{'origx'} or ($w->{'lastindex'} != $i) )
+  {
+   $w->{'origx'} = $bx;
+  }
+
+# Try to keep the same column if possible
+$testX = $w->{'origx'};
+
+# Get the coordinates of the possible new position
+my $testindex = $w->index('@'.$testX.','.$testY );
+$w->see($testindex);
+my ($nx,$ny,$nw,$nh) = $w->bbox($testindex);
+
+# Which side of the character should we position the cursor -
+# mainly for a proportional font
+if ($testX > $nx+$nw/2)
+  {
+   $testX = $nx+$nw+1;
+  }
+
+my $newindex = $w->index('@'.$testX.','.$testY );
+
+if ( $w->compare($newindex,'==','end - 1 char') and ($ny == $ly ) )
+  {
+    # Then we are trying to the 'end' of the text from
+    # the same display line - don't do that
+    return $i;
+  }
+
+$w->{'lastindex'} = $newindex;
+$w->see($newindex);
+return $newindex;
+}
 
 # PrevPara --
 # Returns the index of the beginning of the paragraph just before a given
