@@ -1,8 +1,8 @@
-package Tk::ColorSelect;
+package Tk::ColorSelect; # XXX why is this the Tk::ColorSelect package?
 use strict;
 
 use vars qw($VERSION);
-$VERSION = '4.004'; # $Id: //depot/Tkutf8/Tk/ColorEditor.pm#4 $
+$VERSION = '4.009'; # was: sprintf '4.%03d', q$Revision: #8 $ =~ /\D(\d+)\s*$/;
 
 use Tk qw(Ev);
 
@@ -326,6 +326,7 @@ sub rgbToHsv {
           (($max == $green)?(0.166667*(2 + $rc - $bc)):
            (0.166667*(4 + $gc - $rc)));
     }
+    $hue += 1 if $hue < 0;
     return ($hue, $sat, $max/65535);
 
 } # end rgbToHsv
@@ -442,7 +443,7 @@ sub Show
 package Tk::ColorEditor;
 
 use vars qw($VERSION $SET_PALETTE);
-$VERSION = '4.004'; # $Id: //depot/Tkutf8/Tk/ColorEditor.pm#4 $
+$VERSION = sprintf '4.%03d', q$Revision: #8 $ =~ /\D(\d+)\s*$/;
 
 use Tk qw(lsearch Ev);
 use Tk::Toplevel;
@@ -658,6 +659,7 @@ sub Populate
             sub {
                 my ($objref) = @_;
                 $objref->Callback(-command => ($objref->{'highlight'}, $objref->cget('-color')));
+		$cw->{'done'} = 1;
             }, $cw,
         ],
     );
@@ -711,9 +713,9 @@ sub Populate
 
 sub Show {
 
-    my($objref) = @_;
+    my($objref, @args) = @_;
 
-    $objref->deiconify;
+    Tk::ColorDialog::Show(@_);
 
 } # end show
 
@@ -755,5 +757,3 @@ sub set_colors {
 
 __END__
 
-=cut
-
